
from poseur.constants import MsgEncoding


_INSTALLED_CODECS = {}


def install_codec(new_codec):
    """
    Installs a new codec for use with RPC endpoints.

    If you create your own custom codec, you can install it here with
    the type `MsgEncoding.USER_DEFINED`.

    :param new_codec: The new codec class to install.
    :raises ValueError: If the given codec class is invalid.
    """

    if not issubclass(new_codec, Codec):
        raise ValueError("Given codec to install must be a subclass of Codec")

    if not new_codec.code:
        raise ValueError("Given codec install must have a codec code assigned")

    try:
        new_code = MsgEncoding(new_codec.code)
    except ValueError:
        raise ValueError("Given codec to install has an invalid codec code")

    # Perform Sanity Check
    msg = "Sanity Check"
    encoded = new_codec.encode(msg)
    decoded = new_codec.decode(encoded)
    assert decoded == msg, "{} Failed Sanity Check".format(new_codec.__name__)

    _INSTALLED_CODECS[new_code] = new_codec


def get_codec(code):
    """
    Returns the codec for the given code.

    :param code: The codec code for which the returned codec corresponds to.
    :return: The codec installed for the given code, otherwise `None` if one
     is not installed.
    :raises ValueError: If the given code is an invalid codec code.
    """

    try:
        code = MsgEncoding(code)
    except ValueError:
        raise ValueError("Invalid codec code")

    return _INSTALLED_CODECS.get(code)


class Codec:
    """
    Represents a basic codec used by the RPC endpoints for encoding/decoding
    messages.

    To create your own codec, simply subclass and override the two main
    class methods: `encode` and `decode`.

    All relevant imports for supporting the codec should be done prior to
    codec installation.

    :cvar code: The codec type this codec provides. Must be overridden by all
     subclasses.
    """

    code = None

    @classmethod
    def encode(cls, value):
        raise NotImplementedError("Encode must be implemented by a subclass")

    @classmethod
    def decode(cls, value):
        raise NotImplementedError("Decode must be implemented by a subclass")


class PythonPickleCodec(Codec):
    """
    Codec for Python's Pickle Module
    """

    code = MsgEncoding.PYTHON_PICKLE

    @classmethod
    def encode(cls, value):
        return pickle.dumps(value)

    @classmethod
    def decode(cls, value):
        return pickle.loads(value)


try:
    import pickle
    install_codec(PythonPickleCodec)
except ImportError:                         # pragma: no cover
    pass                                    # pragma: no cover


class MessagePackCodec(Codec):
    """
    Codec for MessagePack.
    """

    code = MsgEncoding.MESSAGE_PACK

    @classmethod
    def encode(cls, value):
        return msgpack.packb(value, use_bin_type=True)

    @classmethod
    def decode(cls, value):
        return msgpack.unpackb(value, use_list=True, encoding='utf-8')


try:
    import msgpack
    install_codec(MessagePackCodec)
except ImportError:                         # pragma: no cover
    pass                                    # pragma: no cover
