
import asyncio
import functools
import traceback
import inspect
import builtins
import uuid

import zmq
import zmq.asyncio
from zmq.auth.asyncio import AsyncioAuthenticator

from poseur.constants import MsgVersion, MsgType, MsgEncoding
from poseur.codecs import get_codec
from poseur import rpc


__all__ = ["Client", "Server"]


# External Loop Setup
zmq.asyncio.install()


class RPCEndpoint:
    """
    Represents an endpoint capable of sending and receiving RPC messages.

    .. note::
        Do not instantiate this class directly, rather use one of its
        subclasses.

    :cvar encoder: Controls the codec used for outgoing messages.
     Override this value to change the encoding scheme used.
    """

    encoder = MsgEncoding.PYTHON_PICKLE

    def __init__(self, rpc_registry=None, loop=None, context=None):
        """
        Initializes the RPC endpoint and prepares the async. loop.

        :param rpc_registry: Override the RPC registry to use.
        :param loop: Override the Async. Loop to use.
        :param context: Override the ZeroMQ context to use.
        """

        self._rpc_registry = rpc_registry or rpc._GLOBAL_RPC_REGISTRY
        self._rpc_futures = {}

        self.__loop = loop or asyncio.get_event_loop()

        self._context = context or zmq.asyncio.Context.instance()

        self._codec = get_codec(self.encoder)
        if not self._codec:
            raise RuntimeError("No codec installed for {}".format(
                str(self.encoder))
            )

        self._socket = self._create_socket()
        self._socket.LINGER = 5
        self._socket.RCVTIMEO = 1000

        def dispatch_reader():
            self.__loop.create_task(self._handler_dispatch())
        self.__loop.add_reader(self._socket.fd, dispatch_reader)

    def _create_socket(self):
        """
        Creates and returns a socket for the specific endpoint type. Must be
         implemented by a subclass.

        :return: A new ZeroMQ socket for use by this endpoint.
        """
        raise NotImplementedError(
            "Socket creation method must be implemented")    # pragma: no cover

    async def _handler_dispatch(self):
        raise NotImplementedError(
            "Method must be implemented by a subclass")      # pragma: no cover

    def _ensure_future_for(self, coro):
        return asyncio.ensure_future(coro, loop=self.__loop)

    def _build_msg(self, msg_type, msg_body=None, msg_version=MsgVersion.V1):
        """
        Builds a Poseur message.

        :returns: The crafted message as a list.
        """
        msg = [MsgVersion(msg_version), MsgType(msg_type)]
        if not msg_body:
            msg.append(MsgEncoding.NONE)
        else:
            msg.append(MsgEncoding(self.encoder))
            msg.append(self._codec.encode(msg_body))
        return msg

    def _build_msg_reg(self):
        """
        Builds a REG Poseur message.

        :returns: The crafted message as a list.
        """
        return self._build_msg(MsgType.REG)

    def _build_msg_task(self, task_id, callable_name, *callable_args,
                        **callable_kwargs):
        """
        Builds a TASK Poseur message.

        :returns: The crafted message as a list.
        """
        if isinstance(task_id, uuid.UUID):
            task_id = task_id.bytes
        body = {
            "task_id": task_id,
            "name": callable_name,
            "args": callable_args,
            "kwargs": callable_kwargs
        }
        return self._build_msg(MsgType.TASK, body)

    def _build_msg_task_result(self, task_id, task_result):
        """
        Builds a TASK_RESULT Poseur message.

        :returns: The crafted message as a list.
        """
        if isinstance(task_id, uuid.UUID):
            task_id = task_id.bytes
        body = {
            "task_id": task_id,
            "task_result": task_result
        }
        return self._build_msg(MsgType.TASK_RESULT, body)

    def _build_msg_task_exception(self, task_id, exc, include_tb=True):
        """
        Builds a TASK_EXCEPTION Poseur message.

        :returns: The crafted message message as a list.
        """
        if isinstance(task_id, uuid.UUID):
            task_id = task_id.bytes
        exc_proxy = traceback.TracebackException.from_exception(exc)
        body = {
            "task_id": task_id,
            "exc_type": exc_proxy.exc_type.__name__,
            "exc_msg": str(exc)
        }
        if include_tb:
            body['exc_tb'] = ''.join(exc_proxy.format())
        return self._build_msg(MsgType.TASK_EXCEPTION, body)

    async def _handle_msg_task(self, msg_body):
        name = msg_body['name']
        task = msg_body['task_id']
        args = msg_body['args']
        kwargs = msg_body['kwargs']
        rpc_callable = self._rpc_registry.get(name)

        if not rpc_callable:
            e = AttributeError("No callable '{}' registered".format(name))
            resp = self._build_msg_task_exception(task, e, include_tb=False)
        else:
            try:
                if inspect.iscoroutinefunction(rpc_callable):
                    retval = await rpc_callable(*args, **kwargs)
                else:
                    retval = rpc_callable(*args, **kwargs)
            except Exception as e:
                # Remove the traceback from this method, to better hide
                # internal implementation details.
                if e.__traceback__:
                    e.with_traceback(e.__traceback__.tb_next)
                resp = self._build_msg_task_exception(task, e)
            else:
                resp = self._build_msg_task_result(task, retval)
        return resp

    async def _handle_msg_task_result(self, msg_body):
        future = self._rpc_futures.pop(msg_body['task_id'])
        if not future:
            print("Warning: No Future Found for Task Result")
            return
        future.set_result(msg_body['task_result'])

    async def _handle_msg_task_exception(self, msg_body):
        future = self._rpc_futures.pop(msg_body['task_id'], None)
        if not future:
            print("Warning: No Future Found for Task Result")
            return
        exc_cls = getattr(builtins, msg_body['exc_type'], Exception)
        exc = exc_cls(msg_body['exc_msg'])
        exc.remote_tb = None
        if 'exc_tb' in msg_body:
            exc.remote_tb = msg_body['exc_tb']
        future.set_exception(exc)

    def start(self):
        """
        Starts the RPC Endpoint and blocks until stopped.
        """
        self.__loop.run_forever()

    def stop(self):
        """
        Stops the RPC Endpoint.
        """
        self.__loop.stop()


class Client(RPCEndpoint):
    """
    A Client is an endpoint capable of connecting to a Server. This `Client`
     is capable of NULL authentication.

    To create and start the client:

    .. code:: python
        client = Client("my-client-username")
        client.connect("tcp://127.0.0.1:1234")  # Connect using ZeroMQ Syntax
        client.start()                          # Blocks
    """

    def __init__(self, username, *args, **kwargs):
        """
        Initializes a new RPC client endpoint capable of NULL authentication.

        :param username: A unique name for this client. Used for authentication
         and identification purposes. This is the name the `Server` will
         use to send messages to the client.
        """

        # (1) Sanitize Input
        # username
        if isinstance(username, str):
            username = username.encode()

        if not isinstance(username, bytes):
            if hasattr(username, "bytes"):
                try:
                    username = username.bytes
                except ValueError:
                    pass
            else:
                try:
                    username = bytes(username)
                except ValueError:
                    pass

        if not isinstance(username, bytes):
            raise ValueError("identity must be byte-like object")

        self.__username = username

        # (2) Initialize Parent
        super().__init__(*args, **kwargs)

    def __getattr__(self, item):
        return functools.partial(self._do_task_on_server, item)

    @property
    def username(self):
        return self.__username

    def _create_socket(self):
        socket = self._context.socket(zmq.DEALER)
        socket.identity = self.username
        return socket

    async def _send_msg(self, msg):
        await self._socket.send_multipart(msg)

    async def _do_task_on_server(self, name, *args, **kwargs):
        # Generate a new Task ID and Task Future
        task_id = uuid.uuid4().bytes
        future = self._rpc_futures[task_id] = asyncio.futures.Future()
        # Send the TASK message and wait for the response.
        task_msg = self._build_msg_task(task_id, name, *args, **kwargs)
        await self._send_msg(task_msg)
        return await future

    async def _handler_dispatch(self):
        """
        Parse and dispatch a handler for the incoming message.
        This method handles incoming messages from DEALER sockets.
        """
        try:
            msg = await self._socket.recv_multipart()
        except zmq.ZMQError:
            return

        # Not enough parts to be valid
        if len(msg) < 3:
            # XXX: Change to debug
            print("Malformed message: Not enough parts")
            return

        try:
            MsgVersion(msg[0])
            p_type = MsgType(msg[1])
            p_encoding = MsgEncoding(msg[2])
        except ValueError as e:
            # XXX: Change to debug
            print("Malformed message: {}".format(e))
            return

        if p_encoding is not MsgEncoding.NONE:
            codec = get_codec(p_encoding)
            if codec:
                try:
                    p_body = codec.decode(msg[3])
                except IndexError:
                    print("Malformed message: Not enough parts")
                    return
                except Exception as e:
                    print("Decoding failed: {}".format(e))
                    return
            else:
                print("No registered decoder for {}".format(p_encoding))
                return
        else:
            p_body = None

        resp = None
        if p_type is MsgType.TASK:
            resp = await self._handle_msg_task(p_body)
        elif p_type is MsgType.TASK_RESULT:
            resp = await self._handle_msg_task_result(p_body)
        elif p_type is MsgType.TASK_EXCEPTION:
            resp = await self._handle_msg_task_exception(p_body)

        # Send a response if one was returned
        if resp:
            await self._send_msg(resp)

    def connect(self, address):
        """
        Connects the RPC Endpoint to the given address.
        """
        try:
            self._socket.connect(address)
        except zmq.ZMQError as e:
            raise ValueError('Invalid address') from e

        # Send a registration message to the server so the connection
        # is logged. Since the loop might not yet be started, we must
        # manually schedule the task on the loop.
        self._ensure_future_for(
            self._send_msg(self._build_msg_reg())
        )


class Server(RPCEndpoint):
    """
    A Server is an endpoint which Clients are capable of connecting to.

    To create and start a server endpoint:

    .. code:: python
        server = Server()
        server.bind("tcp://127.0.0.1:1234")     # Bind using ZeroMQ Syntax
        server.start()                          # Blocks

    :cvar __zap_auth: Handle to the ZAP Authenticator. Only started once per
     process, not per Server instance.
    """

    __zap_auth = None

    @classmethod
    def __start_zap_auth(cls):
        if cls.__zap_auth is None:
            cls.__zap_auth = AsyncioAuthenticator()
            cls.__zap_auth.start()

    def __init__(self, *args, **kwargs):
        """
        Initializes a new RPC server endpoint with ZAP authentication.
        """
        super().__init__(*args, **kwargs)
        self.__start_zap_auth()

    def _create_socket(self):
        socket = self._context.socket(zmq.ROUTER)
        # Emits error on send if no prefix match.
        socket.ROUTER_MANDATORY = True
        # Allows a second connection with a pre-existing prefix to take over
        # the original connection.
        socket.ROUTER_HANDOVER = True
        return socket

    async def _send_msg_to(self, identity, msg):
        # Sanitize incoming identity
        if isinstance(identity, str):
            identity = identity.encode()

        if not isinstance(identity, bytes):
            if hasattr(identity, "bytes"):
                try:
                    identity = identity.bytes
                except ValueError:
                    pass
            else:
                try:
                    identity = bytes(identity)
                except ValueError:
                    pass

        if not isinstance(identity, bytes):
            raise ValueError("identity must be byte-like object")

        msg.insert(0, identity)
        await self._socket.send_multipart(msg)

    async def _do_work_on_client(self, identity, name, *args, **kwargs):
        # Generate a new Task ID and Task Future
        task_id = uuid.uuid4().bytes
        future = self._rpc_futures[task_id] = asyncio.futures.Future()
        # Send the TASK message and wait for the response.
        task_msg = self._build_msg_task(task_id, name, *args, **kwargs)
        await self._send_msg_to(identity, task_msg)
        return await future

    def client(self, username):
        work_func = self._do_work_on_client

        class ClientProxy:
            def __getattr__(self, item):
                return functools.partial(work_func, username, item)

        return ClientProxy()

    async def _handler_dispatch(self):
        """
        Parse and dispatch a handler for the incoming message.
        This method handles incoming messages from ROUTER sockets.
        """
        try:
            msg = await self._socket.recv_multipart()
        except zmq.ZMQError:
            return

        # Not enough parts to be valid
        if len(msg) < 4:
            # XXX: Change to debug
            print("Malformed message: Not enough parts")
            return

        identity = msg[0]

        try:
            MsgVersion(msg[1])
            p_type = MsgType(msg[2])
            p_encoding = MsgEncoding(msg[3])
        except ValueError as e:
            # XXX: Change to debug
            print("Malformed message: {}".format(e))
            return

        if p_encoding is not MsgEncoding.NONE:
            codec = get_codec(p_encoding)
            if codec:
                try:
                    p_body = codec.decode(msg[4])
                except IndexError:
                    print("Malformed message: Not enough parts")
                    return
                except Exception as e:
                    print("Decoding failed: {}".format(e))
                    return
            else:
                print("No registered decoder for {}".format(p_encoding))
                return
        else:
            p_body = None

        resp = None
        if p_type is MsgType.TASK:
            resp = await self._handle_msg_task(p_body)
        elif p_type is MsgType.TASK_RESULT:
            resp = await self._handle_msg_task_result(p_body)
        elif p_type is MsgType.TASK_EXCEPTION:
            resp = await self._handle_msg_task_exception(p_body)

        # Send a response if one was returned
        if resp:
            await self._send_msg_to(identity, resp)

    def bind(self, address):
        """
        Bind the RPC Endpoint to the given address.
        """
        try:
            self._socket.bind(address)
        except zmq.ZMQError as e:
            raise ValueError('Invalid address') from e
